/************************************************************************
 *                                                                      *
 *  control.c  --  interface between rt_control.c and xcontrol.c        *
 *                                                                      *
 *  This file is part of the control/rt_control/xcontrol package which  *
 *  is free software; you can redistribute it and/or modify it under    *
 *  the terms of the version 2 of GNU General Public License as         *
 *  published by the Free Software Foundation.                          *
 *                                                                      *
 *  This program is distributed in the hope that it will be useful,     *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 *  GNU General Public License for more details.                        *
 *                                                                      *
 *  You should have received a copy of the GNU General Public License   *
 *  along with the control/rt_control/xcontrol package; if not, write   *
 *  to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,     *
 *  MA 02139, USA.                                                      *
 *                                                                      *
 *  Copyleft 1999                                                       *
 *  Peter Wurmsdobler                                                   *
 *                                                                      *
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
#include <time.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <mbuff.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gtkplot.h>
#include "common.h"
#include "fifos.h"
#include "shm.h"
#include "dialog.h"
#include "measure.h"
#include "control.h"
#include "xcontrol.h"

/*
 *	Debug macros for 
 */
//#define CONTROL_DEBUG
#ifdef CONTROL_DEBUG
	#define DPRINTF0(message)		printf(message)
	#define DPRINTF1(message,a)		printf(message,a)
	#define DPRINTF2(message,a,b)	printf(message,a,b)
	#define DPRINTF3(message,a,b,c)	printf(message,a,b,c)
#else
	#define DPRINTF0(message)
	#define DPRINTF1(message,a)
	#define DPRINTF2(message,a,b)
	#define DPRINTF3(message,a,b,c)
#endif


/*
 *	Inputs/Outputs to DAQ board defaults, parameters
 *	are set on board and have to be changed here accordingly.
 */
#define DAC_BIT_RANGE		(65536.0) /* for 16 bit */
#define ADC_BIT_RANGE		(65536.0) /* for 16 bit */

#define W_MIN_VOLTAGE		(-10.0)	/* [V] */
#define W_MAX_VOLTAGE		( 10.0)	/* [V] */

#define Y_MIN_VOLTAGE		(-10.0)	/* [V] */
#define Y_MAX_VOLTAGE		( 10.0)	/* [V] */

#define U_MIN_VOLTAGE		(-10.0)	/* [V] */
#define U_MAX_VOLTAGE		( 10.0)	/* [V] */

#define SCALE     ( 1000 ) /* scaling factor for control parameters */


/*
 *	Initialise variables for shared memory
 */	
static shm_t *shm;


/*
 *	Initialise variables for fifos
 */	
static int control_fd;
static int event_fd;

/*	
 *	Send control messages defined in fifos.h to rt_control
 */
inline void control_send_message( unsigned char message )
	{
	DPRINTF0("Starting control_send_message\n");	
	write( control_fd, &message, 1 );
	DPRINTF0("Leaving control_send_message\n");		
    }


/*	
 *	control speed set used in Xcontrol_window.c
 */
int control_set_w( void )
	{
	DPRINTF0("Starting control_set_w\n");	
    MINMAX( W_MIN_VOLTAGE, w, W_MAX_VOLTAGE );
	shm->w = (unsigned short int) ( ( w - W_MIN_VOLTAGE ) / 
    	( W_MAX_VOLTAGE - W_MIN_VOLTAGE ) * DAC_BIT_RANGE );
    DPRINTF1("shm->w = %d\n", shm->w );
    control_send_message( TRIGGER_MEASURE );
    DPRINTF0("Leaving control_set_w\n");	
    return 0;
    }


/*	
 *	control amplitude set used in Xcontrol_window.c
 */
int control_set_N( void )
	{
	DPRINTF0("Starting control_set_N\n");	
    MINMAX( 1, N, SAMPLES );
	shm->N = (unsigned int) ( N );
    DPRINTF1("shm->N = %d\n", shm->N );
	DPRINTF0("Leaving control_set_N\n");	
    return 0;
    }


/*	
 *	control phase set used in Xcontrol_window.c
 */
int control_set_ab( void )
	{
	int l;
    
	DPRINTF0("Starting control_set_ab\n");	
    for (l=0; l<LENGTH; l++)
    	{
   	    shm->a[l] = (int) ( a[l] * SCALE );
	    shm->b[l] = (int) ( b[l] * SCALE );
	    DPRINTF2("b[%d] = %f  ", l, b[l] );
	    DPRINTF2("a[%d] = %f\n", l, a[l] );
	    DPRINTF2("shm->b[%d] = %d  ", l, shm->b[l] );
	    DPRINTF2("shm->a[%d] = %d\n", l, shm->a[l] );
        }
	DPRINTF0("Leaving control_set_ab\n");	
    return 0;
    }


/*	
 *	control measure start
 */
void control_start( void )
	{
	DPRINTF0("Starting control_start\n");	
    control_set_w();
    control_set_N();
    control_set_ab();
    control_send_message( START_CONTROL );
	DPRINTF0("Leaving control_start\n");	
    }


/*	
 *	control measure stop
 */
void control_stop( void )
	{
	DPRINTF0("Starting control_stop\n");	
    control_send_message( STOP_CONTROL );
	DPRINTF0("Leaving control_stop\n");	
    }



static void control_get_measure( void )
	{
    int index;
    gfloat dx,x;
    
	DPRINTF0("Starting control_get_measure\n");	
    x = 0;
    dx = TS/1000.0; /* in ms */
    for ( index=0; index<(shm->N); index++ )
    	{
        y[index].x = x;
        y[index].y = ( ( (gfloat) 
			(shm->y[index]) ) / ADC_BIT_RANGE )	
		    * ( Y_MAX_VOLTAGE - Y_MIN_VOLTAGE )
			+   Y_MIN_VOLTAGE;
		DPRINTF3("y[%d] = [%2.3f, %2.3f], ", index, 
        	(float)(y[index].x), (float)(y[index].y) );	
        
        u[index].x = x;
        u[index].y = ( ( (gfloat) 
			(shm->u[index]) ) / ADC_BIT_RANGE )	
		    * ( U_MAX_VOLTAGE - U_MIN_VOLTAGE )
			+   U_MIN_VOLTAGE;
		DPRINTF3("u[%d] = [%2.3f, %2.3f]\n", index, 
        	(float)(u[index].x), (float)(u[index].y) );	
        
        x += dx;
        }
	DPRINTF0("Leaving control_get_measure\n");	
    }
    

/************************************************************************
 *   All Xcontrol_init related functions                                    *
 ************************************************************************/

/*	
 *	initialize shared memory using Tomasz Motylevskys mbuff device
 */
static void control_init_shm( void )
	{
	DPRINTF0("Starting control_init_shm\n");	
	/* Initialise shared memory between control and rt_control
	 */
	shm = (shm_t *) mbuff_alloc( SHM_NAME, SHM_SIZE );
    
	if ( shm == NULL )
		{
		printf( "Error assigning shared memory at address %p\n", shm );
		printf( "kernel shared memory is not loaded \n" );
		exit(1);
		}
	DPRINTF0("Leaving control_init_shm\n");	
	}


/*	
 *	release shared memory 
 */
static void control_cleanup_shm( void )
	{
	DPRINTF0("Starting control_cleanup_shm\n");	
	mbuff_free( SHM_NAME, (void *) shm );
	DPRINTF0("Leaving control_cleanup_shm\n");	
	}


/*	
 *	initialize fifos from rt module
 */
static void control_init_fifos( void )
	{
	DPRINTF0("Starting control_init_fifos\n");	
	/* Initialise the control message fifo to rt_control
	 */
	control_fd = open( CONTROL_FILE, O_WRONLY );
	if ( control_fd  < 0 )
		{
		printf( "Error opening %s \n", CONTROL_FILE );
		exit(1);
		}

	/* Initialise the event message fifo from rt_control
	 */
	event_fd = open( EVENT_FILE, O_RDONLY );
	if ( event_fd  < 0 )
		{
		printf( "Error opening %s \n", EVENT_FILE );
		exit(1);
		}

	DPRINTF0("Leaving control_init_fifos\n");	
	}


/*	
 *	cleanup fifos from rt module 
 */
static void control_cleanup_fifos( void )
	{
	DPRINTF0("Starting control_cleanup_fifos\n");	
	close(control_fd);
	close(event_fd);
	DPRINTF0("Leaving control_cleanup_fifos\n");	
	}
	

/*	
 *	define event fifo handler for messages passed from
 *	the real time module.
 */
static void control_event_fifo_handler(gpointer data,
	gint source, GdkInputCondition condition)
	{
	unsigned char message;

	DPRINTF0("Starting control_event_fifo_handler\n");	
	
	read( event_fd, &message, 1 );
	
	DPRINTF1("Treating message %c\n",message);
	switch(message)
		{
		case MEASURE_READY:
			DPRINTF0("Treating MEASURE_READY\n");
			control_get_measure();
			measure_show();
			break;
			
		case INVALID_MESSAGE:
			printf("Message %c not valid\n",message);
			break;
		
		default:
			printf("Message %c not possible\n",message);
		}

	DPRINTF0("Leaving control_event_fifo_handler\n");	
	}


/*	
 *	initialize event message handler for GTK
 */
static void control_init_handler( void )
	{
	DPRINTF0("Starting control_init_handler\n");	
	gdk_input_add( (gint) (event_fd), GDK_INPUT_READ, 
		(GdkInputFunction) (control_event_fifo_handler), NULL );
	DPRINTF0("Leaving control_init_handler\n");	
	}


/*	
 *	cleanup event message handler for GTK
 */
static void control_cleanup_handler( void )
	{
	DPRINTF0("Starting control_cleanup_handler\n");	
	gdk_input_remove( event_fd );
	DPRINTF0("Leaving control_cleanup_handler\n");	
	}


/*	
 *	initialize  for xcontrol.c
 */
void control_init( void )
	{
	DPRINTF0("Starting control_init\n");	
	control_init_shm();
	control_init_fifos();
	control_init_handler();
	DPRINTF0("Leaving control_init\n");	
	}


/*	
 *	cleanup for Xcontrol.c
 */
void control_cleanup( void )
	{
	DPRINTF0("Starting control_init\n");	
	control_cleanup_handler();
	control_cleanup_fifos();
	control_cleanup_shm();
	DPRINTF0("Leaving control_init\n");	
	}
