/************************************************************************
 *                                                                      *
 *  rt_bmc1000.c  ---  real time daq input functions.                   *
 *                                                                      *
 *  This file defines all real time measurement input functions used    *
 *  by rt_control.c with functions exported by rt_bmc1000.h. For its    *
 *  parameters it reads the board specific file bmc1000-params.h.       *
 *                                                                      *
 *  This file is part of the control/rt_control/xcontrol package which  *
 *  is free software; you can redistribute it and/or modify it under    *
 *  the terms of the version 2 of GNU General Public License as         *
 *  published by the Free Software Foundation.                          *
 *                                                                      *
 *  This program is distributed in the hope that it will be useful,     *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 *  GNU General Public License for more details.                        *
 *                                                                      *
 *  You should have received a copy of the GNU General Public License   *
 *  along with the control/rt_control/xcontrol package; if not, write   *
 *  to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,     *
 *  MA 02139, USA.                                                      *
 *                                                                      *
 *  Copyleft 1999                                                       *
 *  Peter Wurmsdobler                                                   *
 *                                                                      *
 ************************************************************************/
 
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/kernel.h>
#include <linux/mc146818rtc.h>
#include <linux/ioport.h>
#include <linux/delay.h>
#include "common.h"
#include "rt_bmc1000-params.h"
#include "rt_bmc1000.h"

/*
 *	 Debug macros for DAQ board BMC1000
 */
//#define RT_BMC1000_DEBUG
#ifdef RT_BMC1000_DEBUG
	#define RT_BMC1000_DPRINTF0(message)		printk(message)
	#define RT_BMC1000_DPRINTF1(message,a)		printk(message,a)
	#define RT_BMC1000_DPRINTF2(message,a,b)	printk(message,a,b)
#else
	#define RT_BMC1000_DPRINTF0(message)
	#define RT_BMC1000_DPRINTF1(message,a)
	#define RT_BMC1000_DPRINTF2(message,a,b)
#endif

 
/*
 *	rt_bmc1000_mux_select  --- set the analogue input
 *	range as defined in header file for channel [0-15].
 */
void rt_bmc1000_mux_select(	unsigned int channel,
							unsigned int range_code )
	{
	unsigned char muxbyte;

	/* Generating the byte by channel, offset and 
	 */ 
	muxbyte = (unsigned char) ( ( channel % RT_BMC1000_ADC_CHANNELS )
    	+ 8 );
        
	/* Programming the mux on DAQ board BMC1000.
	 */ 
	outb( muxbyte, RT_BMC1000_MUX_ADDRESS );

	/* We have to keep the CPU busy waiting here for 10us, 
     * because the mux must settle.
	 */
    udelay(10);
	}

/*
 *	rt_bmc1000_aget  ---  get analogue input sample, assumes
 *	that channel is already set with its range.
 */ 
unsigned short int rt_bmc1000_aget( void )
	{
	unsigned char lbyte, hbyte;
    unsigned short int value;

	/* Start conversion by READING from RT_BMC1000_ADC_ADDRESS
	 */
	lbyte = inb( RT_BMC1000_ADC_START_ADDRESS );

	/* We have to keep the CPU busy waiting here for 10us, 
     * which is quite stupid but is due to the board design.
	 */
    udelay(10);

	/* Now we get high and low byte
	 */
	hbyte = inb( RT_BMC1000_ADCH_ADDRESS );
    RT_BMC1000_DPRINTF1("hbyte = 0x%2x\n",hbyte);
	lbyte = inb( RT_BMC1000_ADCL_ADDRESS );	
    RT_BMC1000_DPRINTF1("lbyte = 0x%2x\n",lbyte);

	/* hbyte shifted 8 bits left and add result to lower byte
	 */
	value = (  (unsigned int) ( hbyte << 8 ) + 
 			   (unsigned int) ( lbyte ) );  
	
    return value;
    }


/*
 *	rt_bmc1000_aset  ---  set analogue output values.
 *	Channel must be 0-1.
 */ 
void rt_bmc1000_aset( unsigned short int channel,
					  unsigned short int value )
	{
	unsigned char lbyte, hbyte;
	unsigned int dac_address[RT_BMC1000_DAC_CHANNELS][2] = 
		RT_BMC1000_DAC_ADDRESSES;
	
	/* take 8 last bits as lbyte
	*/
	lbyte = (unsigned char) (value & 0x00ff);

	/* take 8 bits above 8 bits of lbyte and shift 8 bit right
	*/
	hbyte = (unsigned char) ( ( value & 0xff00 ) >> 8 );

	outb( lbyte, dac_address[( channel % RT_BMC1000_DAC_CHANNELS )][0] );
    RT_BMC1000_DPRINTF2("writing hbyte = 0x%2x to 0x%3x\n",
    	hbyte, dac_address[( channel % RT_BMC1000_DAC_CHANNELS )][0] );

	outb( hbyte, dac_address[( channel % RT_BMC1000_DAC_CHANNELS )][1] );
    RT_BMC1000_DPRINTF2("writing lbyte = 0x%2x to 0x%3x\n",
    	lbyte, dac_address[( channel % RT_BMC1000_DAC_CHANNELS )][1] );
	}


/*
 *	rt_bmc1000_init  --- initialises the DAQ board,
 *	but is called form init_module of rt_control.
 *	printk is used only for insmod error messages.
 */
int rt_bmc1000_init( void )
	{
	unsigned int channel;

	RT_BMC1000_DPRINTF0( "Starting rt_bmc1000_init\n" );
		
	/* Request I/O ports for DAQ board BMC1000, base address 
	 * RT_BMC1000_BASE_ADDRESS has to be set via jumpers on baord
	 */
	RT_BMC1000_DPRINTF0( "Requesting I/O ports \n" );
	request_region( RT_BMC1000_BASE_ADDRESS, RT_BMC1000_ADDRESS_RANGE, 
    	RT_BMC1000_ID );
	
	/* Setting analogue input range to a default state
	 */
	RT_BMC1000_DPRINTF0( "Setting analogue output range\n" );
	rt_bmc1000_mux_select(  0, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select(  1, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select(  2, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select(  3, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select(  4, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select(  5, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select(  6, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select(  7, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select(  8, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select(  9, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select( 10, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select( 11, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select( 12, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select( 13, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select( 14, RT_BMC1000_ADC_PM10 );
	rt_bmc1000_mux_select( 15, RT_BMC1000_ADC_PM10 );

	/* Setting DACs to 0 Volts given a range of +-10 Volts
	 */
	RT_BMC1000_DPRINTF0( "Setting all DACs to 32535\n" );
	for (channel=0; channel<RT_BMC1000_DAC_CHANNELS; ++channel)
		rt_bmc1000_aset( channel, 32535 );
        	
	return 0;
	} 

	 
/*
 *	rt_bmc1000_release  --- releases the DAQ board,
 *	but is called form init_module of rt_control.
 */
void rt_bmc1000_release( void )
	{	 
	unsigned int channel;

	RT_BMC1000_DPRINTF0( "Starting rt_bmc1000_release\n" );
	
	/* Setting DACs to 0 Volts given a range of +-10 Volts
	 */
	RT_BMC1000_DPRINTF0( "Setting all DACs to 32535\n" );
	for (channel=0; channel<RT_BMC1000_DAC_CHANNELS; channel++ )
		rt_bmc1000_aset( channel, 32535 );

	/* Release I/O ports from DAQ board BMC1000
	 */
	RT_BMC1000_DPRINTF0( "Releasing I/O ports \n" );
	release_region( RT_BMC1000_BASE_ADDRESS, RT_BMC1000_ADDRESS_RANGE );

	RT_BMC1000_DPRINTF0( "Leaving rt_bmc1000_release\n" );
	} 
