/************************************************************************
 *                                                                      *
 *  rt_control.c  ---   real time control module.                       *
 *                                                                      *
 *  This is the Linux real time module which collects data from data    *
 *  aquisistion boards and passes them to shm. Control is done by       *
 *  a fifo buffer and vice versa.                                       *
 *                                                                      *
 *  This file is part of the control/rt_control/xcontrol package which  *
 *  is free software; you can redistribute it and/or modify it under    *
 *  the terms of the version 2 of GNU General Public License as         *
 *  published by the Free Software Foundation.                          *
 *                                                                      *
 *  This program is distributed in the hope that it will be useful,     *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of      *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       *
 *  GNU General Public License for more details.                        *
 *                                                                      *
 *  You should have received a copy of the GNU General Public License   *
 *  along with the control/rt_control/xcontrol package; if not, write   *
 *  to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge,     *
 *  MA 02139, USA.                                                      *
 *                                                                      *
 *  Copyleft 1999                                                       *
 *  Peter Wurmsdobler                                                   *
 *                                                                      *
 ************************************************************************/
 
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/mc146818rtc.h>
#include <linux/errno.h>
#include <linux/time.h>
#include <linux/mm.h>
#include <rtl_fifo.h>
#include <rtl_time.h>
#include <rtl_sched.h>
#include <mbuff.h>
#include "common.h"
#include "rt_bmc1000.h"
#include "fifos.h"
#include "shm.h"

/*
 *  Debug macros for rt_control
 */
//#define RT_FILTER_DEBUG
#ifdef RT_FILTER_DEBUG
    #define RT_FILTER_DPRINTF0(message)          printk(message)
    #define RT_FILTER_DPRINTF1(message,a)        printk(message,a)
    #define RT_FILTER_DPRINTF2(message,a,b)      printk(message,a,b)
#else
    #define RT_FILTER_DPRINTF0(message)
    #define RT_FILTER_DPRINTF1(message,a)
    #define RT_FILTER_DPRINTF2(message,a,b)
#endif

#define RT_FILTER_VERSION    0x000001
#define RT_FILTER_ID         ("rt_control: ")

#define INPUT_CHANNEL        (0)
#define OUTPUT_CHANNEL       (0)

/*    
 *  Global variables
 */
pthread_t control;           /* THE control thread, or task */
static int measure;          /* flag if measurment is on */
static shm_t *shm;           /* shared memory (sic!) */
static unsigned int index;   /* index in shared memory */
static int e[LENGTH];        /* control deviation buffer  */
static int u[LENGTH];        /* control signal buffer */
static unsigned int k;       /* index k of actual value */
static unsigned short int Y; /* THE measured value, input */
static unsigned short int U; /* THE control current, output */
static unsigned short int W; /* THE setpoint */


/*
 *  Event message handling, send to event message fifo
 *  defined in fifos.h. Function because rtf_put needs
 *  a pointer to the message.
 */
inline void rt_control_event_msg( unsigned char message )
    {                    
    rtf_put( EVENT_FIFO, &message, 1 );
    }


/*
 *  rt_control_thread is started by the scheduler
 *  and carries out the actual control thread.
 */
void *rt_control_thread( void *t )
    {
    unsigned int l,h;
    
    RT_FILTER_DPRINTF0( "Starting rt_control_thread\n" );

    while(1) /* this is the periodic part */
        {
        /* yield to scheduler after init and end of the loop
         */
        RT_FILTER_DPRINTF0( "rt_control_thread, yield to scheduler\n" );

        pthread_wait_np(); 

        /* continue here after being periodically scheduled
         */
        RT_FILTER_DPRINTF0( "rt_control_thread, resuming \n" );
     
        
        /* AND HERE STARTS tTHE CONTROL ALGORITHM */
         
        /* get the value from the board at channel preset in init_module
         */
        Y = rt_bmc1000_aget();    
    
        /* control deviation, 32768 is cancelled out if setpoint
         * is {0,65535}.
         */
        e[k] = (int) ( W - Y ); /* this is the control deviation ! */
    
        /* THE control algorithm, a simple IIR filter
         */
        u[k] = (shm->b[0]) * e[k];
        for ( l=1; l<LENGTH; l++ )
            {
            h = ( LENGTH + k - l ) % LENGTH;
            u[k] += ( (shm->b[l]) * e[h] - (shm->a[l]) * u[h] );
            }
        u[k] /= (shm->a[0]);  
        k = (k+1) % LENGTH;    
    
        /* control variable output, 32768 is added to shift to {0,65535}.
         */
        U = (unsigned short int)( u[k] + 32768 );
        
        /* set the control current on board and the loop is closed
         */
        rt_bmc1000_aset( OUTPUT_CHANNEL, U );
    
        /* if a measurement of step response is set, 
         */
        if (measure == 0)
            {
            /* we store the values in shm
             */
            shm->y[index] = Y;
            shm->u[index] = U;
            
            /* 
             * increase index and stop measurment of step response
              */
            index++;
            if (index == (shm->N) )
                {
                rt_control_event_msg( MEASURE_READY );
                measure = -1;
                }
            } 
        }
    } 



/*
 *    rt_control_start -- start control
 */
static void rt_control_start(void)
    { 
    RT_FILTER_DPRINTF0( "Entering rt_control_start\n" );

    /* cleaning the buffers and reset k  
     */
    for (k=0; k<LENGTH; k++)
        {
        e[k] = 0;
        u[k] = 0;
        }
    k = 0;
     
    /* start the control process by making the thread periodic
     * with the smaple time TS times 1000 to yield ns.
     */ 
    pthread_make_periodic_np( control, gethrtime(), (TS*1000) );
    
    RT_FILTER_DPRINTF0( "Leaving rt_control_start\n" );
    }

/*
 *    rt_control_stop -- stop control.
 */
static void rt_control_stop(void)
    {
    RT_FILTER_DPRINTF0( "Starting rt_control_stop\n" );

    /* stop the control process by setting resume time to
     * infinity and perid to 0, like in thread_create_np.
     */
    pthread_make_periodic_np( control, HRTIME_INFINITY, 0 );
    rt_bmc1000_aset( OUTPUT_CHANNEL, 32768 );

    RT_FILTER_DPRINTF0( "Leaving rt_control_stop\n" );
    }

/*
 *    rt_control_measure -- start measurement
 */
static void rt_control_trigger_measure(void)
    {
    RT_FILTER_DPRINTF0( "Entering rt_control_measure\n" );

    /* reset index and enable one shot measure, then set the 
     * set point. This is for traceing step responses.
     */
    index = 0;
    measure = 0;
    W = shm->w;
    
    RT_FILTER_DPRINTF0( "Leaving rt_control_meaasure\n" );
    }


/*
 *    rt_control_message_handler  ---  is responsible for treating
 *    incoming messages issued by the non real time application.
 */
static int rt_control_message_handler(unsigned int fifo)
    {
    unsigned char message;

    RT_FILTER_DPRINTF0( "Starting message handler\n" );

    while( rtf_get( fifo, &message, 1 ) > 0)
        {
        switch(message)
            {                        
            case START_CONTROL:
                RT_FILTER_DPRINTF0( "Calling rt_control_start()\n" );
                rt_control_start();
                break;
            
            case STOP_CONTROL:
                RT_FILTER_DPRINTF0( "Calling rt_control_stop()\n" );
                rt_control_stop();
                break;
                        
            case TRIGGER_MEASURE:
                RT_FILTER_DPRINTF0( "Calling rt_control_trigger_measure()\n" );
                rt_control_trigger_measure();
                break;
                        
            default:
                rt_control_event_msg( INVALID_MESSAGE );
            }
        }

    RT_FILTER_DPRINTF0( "Leaving message handler\n\n" );
    return(0);
    }



/************************************************************************
 *   All linux kernel module related functions                          *
 ************************************************************************/

/*
 *    init_module  ---  the compulsory init function
 *    printk is used only for insmod error messages.
 */
int init_module(void)
    {
    int l;
    pthread_attr_t attr;            /* control thread attributs */
    struct sched_param sched_param; /* control thread scheduler parameter */
        
    RT_FILTER_DPRINTF0( "Starting module init\n" );

    /* Initialise the PCL818 DAQ board and set input channel
     */
    RT_FILTER_DPRINTF0( "Calling rt_pc1000_init function\n" );
    if ( rt_bmc1000_init() )
        {
        printk( "%s Calling rt_pc1000_init failed\n", RT_FILTER_ID );
        return -ENOMEM;
        }
    rt_bmc1000_mux_select(  INPUT_CHANNEL, RT_BMC1000_ADC_PM10 );

    /* Initialise rt-fifos
     */
    if ( rtf_create( CONTROL_FIFO, FIFO_SIZE ) < 0 )
        {
        printk( "%s Could not install fifo %d\n",
            RT_FILTER_ID, CONTROL_FIFO );
        return -ENOMEM;
        }
    if ( rtf_create( EVENT_FIFO, FIFO_SIZE ) < 0 )
        {
        printk( "%s Could not install fifo %d\n",
            RT_FILTER_ID, EVENT_FIFO );
        return -ENOMEM;
        }

    /* Initialise message handler
     */
    RT_FILTER_DPRINTF0( "Installing rt_control_message_handler\n" );
    if ( rtf_create_handler( CONTROL_FIFO, rt_control_message_handler ) )
        {
        printk( "%s Could not install rt_control_message_handler\n",
            RT_FILTER_ID );
        return -ENOMEM;
        }

    /* Initialise shared memory using Tomasz Motylevski's mbuff.o
     */    
    if ( shm_allocate( SHM_NAME, SHM_SIZE, (void **) &shm ) < 0 )
        {
        printk( "%s Initialising shared memory failed\n", RT_FILTER_ID );
        return -ENOMEM;
        }  
    shm->w = 32535;
    shm->N = SAMPLES;
    shm->a[0] = 1;
    shm->b[0] = 1;
    for (l=1; l<LENGTH; l++)
        {
        shm->a[l] = 0;
        shm->b[l] = 0;
        }

    /* and now we initialize the kernel thread, with attributes
     * scheduler parameters and then create it.
     */
    pthread_attr_init( &attr );
    pthread_attr_setcpu_np( &attr, 0 );
    sched_param.sched_priority = 1;
    pthread_attr_setschedparam( &attr, &sched_param );
    if ( pthread_create( &control,  &attr, rt_control_thread, (void *)1 ) )
       {
       printk( "Initializing real time control thread failed" );
       return -ENOMEM;
       }

    printk( "%s Init module sucessfull\n", RT_FILTER_ID );

    RT_FILTER_DPRINTF0( "Leaving module init.\n\n" );
    return(0);
    }
        

/*
 *    cleanup_module  ---  the compulsory cleanup function
 */
void cleanup_module(void)
    {
    RT_FILTER_DPRINTF0( "Starting module cleanup\n" );
        
    /* Delete control thread
     */
    pthread_delete_np( control );
                    
    /* Release shared memory using Tomasz Motylevski's mbuff.o
     */
    shm_deallocate( shm );

    /* Release rt-fifos
     */
    rtf_destroy( CONTROL_FIFO );
    rtf_destroy( EVENT_FIFO );

    /* Release the PC1000 DAQ board
     */
    RT_FILTER_DPRINTF0( "Calling rt_pc1000_release function\n" );
    rt_bmc1000_release();

    printk( "%s Cleanup module sucessfull\n", RT_FILTER_ID );

    RT_FILTER_DPRINTF0( "Leaving module cleanup.\n\n" );
    }
